import numpy as np
import tensorflow_datasets as tfds
import tensorflow as tf
tfds.disable_progress_bar()
from matplotlib import pyplot as plt
import pandas as pd
import os
import sys
from TextClassification.argsparse_classification_preamble import argsparse_preamble
import classification_metrics as cls_metrics
import time
from database_preparation.utils_labeled_datasets import get_splits_for_cross_val
from keras.models import Sequential
from keras.layers import Dense
from keras.layers import Flatten
from keras.layers.convolutional import Conv1D
from keras.layers.convolutional import MaxPooling1D
#from keras.layers.embeddings import Embedding
from tensorflow.keras.layers import Embedding
from keras_preprocessing.sequence import pad_sequences
from database_preparation.utils_labeled_datasets import text_label_files_to_labeled_dataset
from database_preparation.preprocess import print_meta_data
sys.path.append(os.getcwd())

def plot_graphs(history, metric):
    plt.close()
    plt.plot(history.history[metric])
    plt.plot(history.history['val_'+metric])
    plt.xlabel("Epochs")
    plt.ylabel(metric)
    plt.legend([metric, 'val_'+metric])
    plt.show()

def save_graphs(history, metric, save_path):
    plt.close()
    plt.plot(history.history[metric])
    plt.plot(history.history['val_' + metric])
    plt.xlabel("Epochs")
    plt.ylabel(metric)
    plt.legend([metric, 'val_' + metric])
    #plt.show()
    plt.savefig(save_path, dpi=300)
    print("generated "+save_path)

def dict2tf_dataset(dict):
    return tf.data.TextLineDataset.from_tensor_slices(([text for text in dict['text']], dict['label']))

def train_test_updatemetrics(train_dataset, test_dataset, num_classes, metrics,
                             num_epochs=50, plot_loss=False,
                             plot_save_path="TextClassification/plots/CNN/CNN_loss.png"):
    train_dataset = dict2tf_dataset(train_dataset)
    test_dataset = dict2tf_dataset(test_dataset)

    ############### Create the text encoder ###################
    VOCAB_SIZE = 5000
    encoder = tf.keras.layers.TextVectorization(
        max_tokens=VOCAB_SIZE)
    encoder.adapt(train_dataset.map(lambda text, label: text))

    vocab = np.array(encoder.get_vocabulary())

    # encode data to word-indices:
    X_train = []
    y_train = []
    X_test = []
    y_test = []
    for text, label in train_dataset:
        X_train.append(encoder(text).numpy())
        y_train.append(label)
    for text, label in test_dataset:
        X_test.append(encoder(text).numpy())
        y_test.append(label)
    X_train = np.asarray(X_train)
    y_train = np.asarray(y_train)
    X_test = np.asarray(X_test)
    y_test = np.asarray(y_test)

    ##### create the model: #####

    # Padding the data samples to a maximum review length in words
    max_words = 450
    X_train = pad_sequences(X_train, maxlen=max_words)
    X_test = pad_sequences(X_test, maxlen=max_words)

    # Building the CNN Model
    embedding_dim = 100
    filter_amount = 32
    filter_size = 3
    model = Sequential()  # initilaizing the Sequential nature for CNN model
    model.add(Embedding(len(encoder.get_vocabulary()), embedding_dim, input_length=max_words, mask_zero=True))
    model.add(Conv1D(filter_amount, filter_size, padding='same', activation='relu'))
    model.add(MaxPooling1D())
    model.add(Flatten())
    model.add(Dense(250, activation='relu'))
    model.add(Dense(num_classes, activation='softmax'))
    model.compile(loss='sparse_categorical_crossentropy',
                  optimizer=tf.keras.optimizers.Adam(1e-4),
                  metrics=["accuracy"])
    # model.summary()


    start = time.time()

    # evaluate:
    if plot_loss:
        history = model.fit(X_train, y_train, epochs=num_epochs,
                            batch_size=128, verbose=2,
                            validation_data=(X_test, y_test),
                            validation_steps=1,
                            )
        save_graphs(history, 'loss', plot_save_path)

    else:
        model.fit(X_train, y_train, epochs=num_epochs,
                  batch_size=128, verbose=2)
        predictions = model.predict(X_test)
        y_pred = np.argmax(predictions, axis=-1)
        metrics.update_metrics(y_test, y_pred, True, start)

def main():
    ############# get labeled text data ###################

    args = argsparse_preamble()

    print("CNN Evaluation with corpus " + args.path2corpus + " and cluster set " + args.clustered_data)
    print("infos about corpus:")
    print_meta_data(args.path2corpus)

    dataset = text_label_files_to_labeled_dataset(args.clustered_data, path2corpus=args.path2corpus)

    num_classes = int(pd.DataFrame(dataset["label"]).nunique())

    metrics = cls_metrics.ClassificationMetrics("CNN")

    epochs = 100
    folds = 10

    for i, (train_dataset, test_dataset) in enumerate(get_splits_for_cross_val(dataset, folds)):
        if args.loss_curve_check:
            train_test_updatemetrics(train_dataset, test_dataset, num_classes, metrics,
                                     epochs, plot_loss=True,
                                     plot_save_path="TextClassification/plots/CNN/CNN_loss_"+str(i+1)+".png")
        else:
            print("====== CNN train/test run " + str(i+1) + "/" + str(folds) + " ======")
            print(str(len(train_dataset["label"]))+" train documents")
            print(str(len(test_dataset["label"])) + " test documents")
            train_test_updatemetrics(train_dataset, test_dataset, num_classes, metrics, epochs)

    if not args.loss_curve_check:
        metrics.save_scores_to_disk(args.clustered_data)
        metrics.pickle_object(args.clustered_data)
        cls_metrics.print_results_as_latextable(metrics.json_file_path)


if __name__ == '__main__':
    main()