# -*- coding: iso-8859-1 -*-
import os
from sklearn.naive_bayes import MultinomialNB
from sklearn.linear_model import LogisticRegression
from sklearn.linear_model import SGDClassifier  # stochastic gradient descent (SGD) learning
from sklearn.neural_network import MLPClassifier
from sklearn.decomposition import TruncatedSVD
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.pipeline import Pipeline
from TextClassification.argsparse_classification_preamble import argsparse_preamble
import database_preparation.utils_labeled_datasets as dt
from database_preparation.utils_labeled_datasets import get_splits_for_cross_val
import TextClassification.classification_metrics as cls_metrics
from database_preparation.preprocess import print_meta_data
from database_preparation.utils_labeled_datasets import is_text_lst_tfidf_vectorized
import pickle
import numpy as np
import pandas as pd
'''from imblearn.under_sampling import RandomUnderSampler
from imblearn.over_sampling import RandomOverSampler
from imblearn.pipeline import make_pipeline as make_pipeline_imb'''

#expeeriment:
merge_classes = [(0, 1), (5, 7), (9, 10, 11), (6, 15)]

# for tfidf vectorizer
def identity(words):
    return words


def create_pipeline(estimator, reduction=False, with_vectorizer=True):
    '''
    construct a pipeline with sklearn.pipeline
    pased estimator will be the last element of the pipeline
    using tfidf as vectorizer
    '''
    steps = []

    if with_vectorizer:
        steps.append(
            ('vectorize', TfidfVectorizer(tokenizer=identity, preprocessor=None, lowercase=False))
        )

    if reduction:
        steps.append((
            'reduction', TruncatedSVD(n_components=1000)
        ))

    # Add the estimator
    steps.append(('classifier', estimator))
    return Pipeline(steps)


def get_immediate_subdirectories(a_dir):
    return [name for name in os.listdir(a_dir)
            if os.path.isdir(os.path.join(a_dir, name))]


def cross_validate_with_bow_classifiers(label_set, fold_amount=10,
                                        path2corpus="./database/bow_prepro_desc.pkl",
                                        df_cases_file="database/df_cases.pkl"):

    '''
    cross validates passed label_set with text data saved in path2corpus and labels saved in df_cases_file.
    path2corpus should be a list of reports, where each report is tokenized
    or a list of tf-idf vectorized texts (of type scipy.sparse.csr.csr_matrix).
    '''



    if is_text_lst_tfidf_vectorized(path2corpus):
        models = []
        models.append(create_pipeline(MultinomialNB(), with_vectorizer=False))
        models.append(create_pipeline(MLPClassifier(max_iter=300), with_vectorizer=False))
        models.append(create_pipeline(LogisticRegression(), with_vectorizer=False))
        models.append(create_pipeline(SGDClassifier(), with_vectorizer=False))

        with open(path2corpus, 'rb') as f:
            loaded_texts = pickle.load(f)
        n = np.asarray(loaded_texts.todense().tolist())

        from database_preparation.utils_labeled_datasets import text_label_2_labeled_dataset
        df_cases = pd.read_pickle(df_cases_file)
        dataset = text_label_2_labeled_dataset(n, df_cases["label_" + label_set])

    else:
        # create model-pipelines for cross-validation with different pipelines:
        models = []
        models.append(create_pipeline(SGDClassifier()))
        models.append(create_pipeline(MultinomialNB()))
        models.append(create_pipeline(LogisticRegression()))
        models.append(create_pipeline(MLPClassifier(max_iter=300)))
        '''models.append(make_pipeline_imb(TfidfVectorizer(tokenizer=identity, preprocessor=None, lowercase=False)
                                        , RandomOverSampler(), SGDClassifier()))'''

        # print(f"train models {[model['classifier'] for model in models]} with corpus {path2corpus} and cluster set {label_set}")
        print("infos about corpus:")
        print_meta_data(path2corpus)

        dataset = dt.text_label_files_to_labeled_dataset(label_set, path2corpus=path2corpus,
                                                         df_cases_path=df_cases_file)

    # in order to use same 10-fold-cross-splits for each model:
    k_train_test_sets = []
    for (train_dataset, test_dataset) in get_splits_for_cross_val(dataset,
                                            fold_amount, merge_classes=None, oversample=False, stratified=True):
        k_train_test_sets.append(tuple((train_dataset, test_dataset)))

    # cross validate each model and save metrics:
    for model in models:
        print('running ' + str(model['classifier']))
        name = model.named_steps['classifier'].__class__.__name__
        if 'reduction' in model.named_steps:
            name += " (TruncatedSVD)"

        metrics = cls_metrics.ClassificationMetrics(name)

        for i, (train_dataset, test_dataset) in enumerate(k_train_test_sets):
            model.fit(train_dataset['text'], train_dataset['label'])
            y_pred = model.predict(test_dataset['text'])

            metrics.update_metrics(test_dataset['label'], y_pred, False)

        metrics.save_scores_to_disk(label_set)
        metrics.pickle_object(label_set)
        df = metrics.classes_scores(-1)
        print(df.to_latex().replace('{}', 'cluster'))


    cls_metrics.print_results_as_latextable(metrics.json_file_path)



def main():
    args = argsparse_preamble()
    cross_validate_with_bow_classifiers(args.clustered_data, path2corpus=args.path2corpus)


if __name__ == '__main__':
    main()
