
#%% import
from sklearn.metrics import silhouette_score, calinski_harabasz_score, davies_bouldin_score
from validclust import cop, dunn
from sklearn.metrics import pairwise_distances
import pandas as pd
import os

#%% class definition
class ClusterMetrics:

    def __init__(self, feature_matrix, feature_label, file_name = "cluster_metrics.pkl"):

        self.file_name = file_name

        self.feature_matrix = feature_matrix

        if (type(feature_matrix).__name__) == "csr_matrix":
            self.__feature_matrix_array__ = feature_matrix.toarray()
        elif (type(feature_matrix).__name__) == "list":
            self.__feature_matrix_array__ = np.array(feature_matrix)
        else:
           self.__feature_matrix_array__  = feature_matrix

        self.feature_label = feature_label
        self.__feature_label_array = np.array(feature_label)

        self.__distance_matrix__ = pairwise_distances(self.__feature_matrix_array__)

        # calculate Silhouette Coefficien (values -1 to 1)
        self.s_score = silhouette_score(feature_matrix, feature_label)

        # caculate Calinski-Harabasz Index (the higher the value, the better)
        self.ch_index = calinski_harabasz_score(self.__feature_matrix_array__ , feature_label)

        # calcualte the Davies-Bouldin Index (the highter, the better)
        self.db_score = davies_bouldin_score(self.__feature_matrix_array__, feature_label)

        # calculate COP CVI
        self.cop = cop(self.__feature_matrix_array__, self.__distance_matrix__, self.__feature_label_array)

        # calculate Dunn CVI
        self.dunn = dunn(self.__distance_matrix__, self.__feature_label_array)

        # place for entropy
        self.entropy = None

        self.svm_accuracy = None

    def write_to_file(self):

        results = [np.round(self.s_score,3),
                   np.round(self.ch_index,3),
                   np.round(self.db_score,3),
                   np.round(self.cop,3),
                   np.round(self.dunn,3),
                   self.entropy,
                   self.svm_accuracy]

        head, tail = os.path.split(self.file_name)
        tail = tail[:-4]

        df = pd.DataFrame(results,
                          index =['s-score', 'ch-index', 'db-score', 'cop', 'dunn-score', 'entropy', 'svm-accuracy'],
                          columns =[tail])
        df.to_pickle(self.file_name)

        print(df)

    def __str__(self):
        return "s-score: " + str(np.round(self.s_score,2)) + "[-1:1]" + "\n" + \
               "ch-index: " + str(np.round(self.ch_index,2)) + " [0:]" "\n" + \
               "db-score: " + str(np.round(self.db_score, 2)) + "[0:]" "\n" + \
                "cop: " + str(np.round(self.cop, 2)) + " []" "\n" + \
                "dunn: " + str(np.round(self.dunn, 2)) + " []" "\n" + \
               "entropy: " + str(self.entropy) + " []" "\n" + \
               "svm-accuracy: " + str(self.svm_accuracy) + " []"

            #%% define plot functions for PCA
from sklearn.decomposition import PCA
import matplotlib.pyplot as plt

def plot_pca(text_features, labels, file_path = [], show_plot = True):

    pca = PCA(n_components=2)
    reduced_features = pca.fit_transform(text_features)

    plt.close()
    plt.scatter(x=reduced_features[:, 0], y=reduced_features[:, 1],
                c=np.int8(labels), cmap="tab20")
    plt.colorbar()
    plt.title('PCA-representation')

    if bool(file_path):
        plt.savefig(file_path)

    if show_plot:
        plt.show()

#%% define plot function for T-SNE
from sklearn.manifold import TSNE

def plot_tsne(text_features, labels, file_path = [], show_plot = True):

    tsne = TSNE(n_components=2, verbose=1, random_state=123)
    reduced_features = tsne.fit_transform(text_features)

    plt.close()
    plt.scatter(x=reduced_features[:, 0], y=reduced_features[:, 1],
                c=np.int8(labels), cmap="tab20")
    plt.colorbar()
    plt.title('T-SNE-representation')
    if bool(file_path):
        plt.savefig(file_path)
    if show_plot:
        plt.show()

#%% define plot function for UMAP
import umap
import seaborn as sns
def cluster_scatter_plot(umap_text_features2D, labels, file_path = [],
                         show_plot=True, colorblindfriendly=True,
                         number_data_points=True, fig_title=None):

    '''umap_text_features2D = umap.UMAP(n_neighbors=15,
                                     n_components=2,
                                     min_dist=0.0, metric='cosine').fit_transform(text_features)'''

    plt.close()

    # plot unvalid labeled datapoints in black:
    x = [e for i, e in enumerate(umap_text_features2D[:, 0]) if labels[i] == -1]
    y = [e for i, e in enumerate(umap_text_features2D[:, 1]) if labels[i] == -1]
    if len(x) > 0:
        plt.scatter(x=x, y=y, c='black', marker='.')

    # plot valid labeled datapoints:
    x_val = [e for i,e in enumerate(umap_text_features2D[:, 0]) if labels[i] != -1]
    y_val = [e for i,e in enumerate(umap_text_features2D[:, 1]) if labels[i] != -1]

    valid_labels = [l for l in labels if l != -1]
    if colorblindfriendly:# use colorblind palette, it has 10 colors
        style = []
        for label in valid_labels:
            if label <= 9: #0-9
                style.append(0)
            elif label >= 20: #20-inf
                style.append(1)
            else: #10-19
                style.append(2)
        sns.scatterplot(x=x_val, y=y_val, hue=valid_labels,
                        palette="colorblind", style=style,
                        legend=True, linewidth=.3)
        if number_data_points:
            nummerate_clusters_in_plot(x_val, y_val, valid_labels)

    else: # use tap20, it has 20 different colors


        x = [e for i, e in enumerate(x_val) if valid_labels[i] <= 19]
        y = [e for i, e in enumerate(y_val) if valid_labels[i] <= 19]
        c = [e for e in valid_labels if e <= 19]
        plt.scatter(x=x,
                    y=y,
                    c=np.int8(c),
                    cmap="tab20", edgecolors='white', linewidth=.3
                    , marker='o')
        if number_data_points:
            nummerate_clusters_in_plot(x,y,c)
        '''plt.legend(handles=scatter.legend_elements()[0],
                   labels=[str(l) for l in c], loc="best")'''
        plt.colorbar(values=[int(e) for e in np.unique(np.asarray(c))])

        c = [e for e in valid_labels if e > 19]
        if len(c)>0:
            x = [e for i, e in enumerate(x_val) if valid_labels[i] > 19]
            y = [e for i, e in enumerate(y_val) if valid_labels[i] > 19]
            plt.scatter(x=x, y=y,
                        c=np.int8(c),
                        cmap="tab20", edgecolors='white', linewidth=.3
                        , marker='P')
            if number_data_points:
                nummerate_clusters_in_plot(x, y, c)


    if fig_title is None:
        if bool(file_path):
            import os
            fig_title = os.path.basename(file_path)
        else:
            fig_title = "UMAP"

    plt.title(fig_title.replace(".png",""))

    if bool(file_path):
        print("generated "+file_path)
        plt.savefig(file_path,dpi=300)
    if show_plot:
        plt.show()


def nummerate_clusters_in_plot(x,y,labels):
    annotated_labels = []
    for i, label in enumerate(labels):
        if label not in annotated_labels:
            plt.annotate(label, (x[i], y[i]))
            annotated_labels.append(label)

def Jaccard_Similarity(doc1, doc2):

    if isinstance(doc1, list):
        doc1 = " ".join(doc1)
        doc2 = " ".join(doc2)

    # List the unique words in a document
    words_doc1 = set(doc1.lower().split())
    words_doc2 = set(doc2.lower().split())

    #print(words_doc1)
    #print(words_doc2)

    # Find the intersection of words list of doc1 & doc2
    intersection = words_doc1.intersection(words_doc2)

    # Find the union of words list of doc1 & doc2
    union = words_doc1.union(words_doc2)

    # Calculate Jaccard similarity score
    # using length of intersection set divided by length of union set
    return float(len(intersection)) / len(union)

#%%
import numpy as np

def get_distance_matrix(str_list):

    dist_matrix = np.zeros(shape=(len(str_list), len(str_list)))

    # calculate the lower triangle
    for i in range(0, len(str_list)):
        for j in range(i+1, len(str_list)):
                dist_matrix[i][j] = Jaccard_Similarity(str_list[i], str_list[j])
    # fill the upper triangle
    for i in range(0, len(str_list)):
        for j in range(0, len(str_list)):
            if i == j:
                dist_matrix[i][j] = 0
            elif i > j:
                dist_matrix[i][j] = dist_matrix[j][i]

    return dist_matrix
